import json
import logging
import os
import yaml

from .inference import *
from .client import Client
from .server import Server
from .utils import to_unix_path, update_dict_with_unix_paths
from .base_inference import ServingConfig
from .base_inference import PythonConfig
from .base_inference import PythonStep

KONDUIT_BASE_DIR = os.getcwd()
KONDUIT_PID_STORAGE = os.path.join(KONDUIT_BASE_DIR, "pid.json")
MODEL_TYPES = [
    "TENSORFLOW",
    "KERAS",
    "DL4J",
    "ONNX",
    "PMML",
    "SAMEDIFF",
]


def store_pid(file_path, pid):
    """ Store the process ID for a running Konduit Server, given a configuration file that
    the server has been started from.

    :param file_path: path to your Konduit configuration file
    :param pid: process ID of the Konduit server you created before.
    """
    yaml_path = os.path.abspath(file_path)
    pid_dict = {yaml_path: pid}
    with open(KONDUIT_PID_STORAGE, "w") as f:
        try:
            previous = json.load(f)
            new = previous.copy()
            new.update(pid_dict)
            json.dump(new, f)
        except:
            json.dump(pid_dict, f)


def pop_pid(file_path):
    """Remove a process ID from the list of running processes. Use this if you want
    to kill that process immediately after.

    :param file_path: path to the Konduit configuration file that your Konduit server has been created from.
    :return: the process ID belonging to that Konduit server instance.
    """
    pid = None
    with open(KONDUIT_PID_STORAGE, "r+") as f:
        yaml_path = os.path.abspath(file_path)
        try:
            previous = json.load(f)
            if yaml_path in previous:
                pid = previous.pop(yaml_path)
            f.seek(0)
            f.write(json.dumps(previous))
            f.truncate()
        except Exception as e:
            logging.warning("Process ID not found, file is empty.", e)
    return pid


def load_data(file_path, use_yaml=True):
    """Load data from a given YAML file into a Python dictionary.

    :param file_path: path to your YAML file.
    :param use_yaml: use yaml or json
    :return: contents of the file as Python dict.
    """
    with open(file_path, "r") as f:
        if use_yaml:
            data = yaml.safe_load(f)
        else:
            data = json.load(f)
    return data


def pop_data(serving_dictionary, dict_key):
    """Return a value from a Python dictionary for a given key, if it exists,
    and also remove the key-value pair afterwards.
    
    :param serving_dictionary: 
    :param dict_key:
    :return: 
    """
    result = None
    if dict_key in serving_dictionary:
        result = serving_dictionary.pop(dict_key)
    return result


def from_file(file_path, start_server=True, use_yaml=True):
    """Create Konduit Server and Client from file

    :param file_path: path to your konduit.yaml
    :param start_server: whether to start the server instance or not (if not you can start it later). Defaults to
           True so that the client can be loaded successfully.
    :param use_yaml: use yaml or json
    :return: konduit.server.Server and konduit.client.Client instances
    """
    return (
        server_from_file(file_path, start_server, use_yaml),
        client_from_file(file_path, use_yaml),
    )


def server_from_file(file_path, start_server=False, use_yaml=True):
    """Create a Konduit Server from a from a configuration file.

    :param file_path: path to your konduit.yaml
    :param start_server: whether to start the server instance or not (if not you can start it later). Defaults to
           False, meaning you have to start the server manually by calling "start()" explicitly.
    :param use_yaml: use yaml or json
    :return: konduit.server.Server instance
    """
    file_path = to_unix_path(file_path)
    data = load_data(file_path, use_yaml)
    serving_data = data.get("serving", None)

    extra_start_args = pop_data(serving_data, "extra_start_args")
    extra_jar_args = pop_data(serving_data, "extra_jar_args")
    jar_path = pop_data(serving_data, "jar_path")
    config_path = pop_data(serving_data, "config_path")
    if not config_path:
        config_path = "config.json"

    serving_config = ServingConfig(**serving_data)

    step_data = data.get("steps", None)
    steps = []
    for step_config in step_data:
        steps.append(get_step(step_config))

    server = Server(
        serving_config=serving_config,
        steps=steps,
        extra_start_args=extra_start_args,
        extra_jar_args=extra_jar_args,
        jar_path=jar_path,
        config_path=config_path,
    )
    if start_server:
        server.start()
    return server


def client_from_file(file_path, use_yaml=True):
    """Create a Konduit client instance from a configuration file.
    If your konduit configuration file has a "client" section, that
    is used to create the client instance. If it doesn't, all properties
    are derived from the "serving" section.

    :param file_path: path to your konduit.yaml
    :param use_yaml: use yaml or json
    :return: konduit.client.Client instance
    """
    file_path = to_unix_path(file_path)
    data = load_data(file_path, use_yaml)
    if "client" in data.keys():
        client_data = data.get("client", None)
        client = Client(**client_data)
    else:
        client_data = data.get("serving", None)
        port = client_data.get("http_port", None)
        input_data_format = client_data.get("input_data_format", None)
        output_data_format = client_data.get("output_data_format", None)

        if not port:
            raise RuntimeError(
                "No HTTP port found in configuration file, can't proceed."
            )
        client = Client(
            port=port,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
        )
    return client


def get_step(step_config):
    """Get a PipelineStep from a step configuration.

    :param step_config: python dictionary with properties to create a PipelineStep
    :return: konduit.inference.PipelineStep instance.
    """
    step_type = step_config.pop("type")
    step_config = update_dict_with_unix_paths(step_config)

    if step_type == "PYTHON":
        step = get_python_step(step_config)
    elif step_type in MODEL_TYPES:
        step = get_model_step(step_config, step_type)
    elif step_type == 'IMAGE':
        step = get_image_load_step(step_config)
    elif step_type == "WORDPIECE_TOKENIZER":
        step = get_wordpiece_tokenizer_step(step_config)
    else:
        raise Exception("Step type of type " + step_type + " currently not supported.")
    return step


def get_python_step(step_config):
    """Get a PythonStep from a configuration object

    :param step_config: python dictionary with properties to create a PipelineStep
    :return: konduit.inference.PythonStep instance.
    """

    python_step = PythonStep()
    for key, config in step_config["python_configs"].items():
        python_step.step(python_config=PythonConfig(**config), input_name=key)

    return python_step


def get_image_load_step(step_config):
    """Get a ImageLoadingStep from a configuration object

    :param step_config: python dictionary with properties to create the ImageLoadingStep
    :return: konduit.inference.ImageLoadingStep instance.
    """
    step = ImageLoadingStep(**step_config)
    return step


def get_wordpiece_tokenizer_step(step_config):
    """Get a BertStep from a configuration object

    :param step_config: python dictionary with properties to create the BertStep
    :return: konduit.inference.BertStep instance.
    """
    step = WordPieceTokenizerStep(**step_config)
    return step


def get_model_step(step_config, step_type):
    """Get a ModelStep from a configuration object

    :param step_config: python dictionary with properties to create a PipelineStep
    :param step_type: type of the step (TENSORFLOW, KERAS, DL4J, PMML or SAMEDIFF)
    :return: konduit.inference.ModelStep instance.
    """

    if (
        step_type == "TENSORFLOW"
    ):  # TF has to extra properties, all others are identical
        model_step = TensorFlowStep(**step_config)
    elif (
            step_type == "KERAS"
    ):  # TF has to extra properties, all others are identical
        model_step = KerasStep(**step_config)
    elif (
            step_type == "DL4J"
    ):  # TF has to extra properties, all others are identical
        model_step = Dl4jStep(**step_config)
    elif (
            step_type == "SAMEDIFF"
    ):  # TF has to extra properties, all others are identical
        model_step = SameDiffStep(**step_config)
    elif (
            step_type == "PMML"
    ):  # TF has to extra properties, all others are identical
        model_step = PmmlStep(**step_config)
    else:
        model_step = ModelStep(**step_config)

    return model_step



import numpy as np
import pytest
from konduit import *
from konduit.server import Server
from konduit.utils import is_port_in_use


@pytest.mark.integration
def test_setup_and_run_start():
    python_config = PythonConfig(
        python_code="def setup(): pass\ndef run(input): {'output': np.array(input + 2)}",
        python_inputs={"input": "NDARRAY"},
        python_outputs={"output": "NDARRAY"},
        setup_and_run=True,
    )

    step = PythonStep().step(python_config)
    server = Server(steps=step, serving_config=ServingConfig())
    _, port, started = server.start()

    assert started
    assert is_port_in_use(port)

    client = server.get_client()

    data_input = {"default": np.asarray([42.0, 1.0])}

    try:
        predicted = client.predict(data_input)
        print(predicted)
        server.stop()
    except Exception as e:
        print(e)
        server.stop()



import numpy as np
import pytest
from konduit.load import server_from_file, client_from_file


@pytest.mark.integration
def test_yaml_server_python_prediction():
    try:
        konduit_yaml_path = "yaml/konduit_tf_inference.yaml"
        img = np.load("../data/input_layer.npy")
        server = server_from_file(konduit_yaml_path, start_server=True)
        client = client_from_file(konduit_yaml_path)
        predicted = client.predict(data_input={"input_layer": img})
        result = dict(zip(np.arange(10), predicted[0].round(3)))
        # {0: 0.0, 1: 0.0, 2: 0.001, 3: 0.001, 4: 0.0, 5: 0.0, 6: 0.0, 7: 0.998, 8: 0.0, 9: 0.0}
        assert round(result.get(7) * 1000) == 998
        server.stop()
    finally:
        server.stop()



import konduit
import numpy as np
import pytest
from konduit.load import server_from_file, client_from_file


@pytest.mark.integration
def test_yaml_server_loading():
    file_path = "yaml/konduit.yaml"
    server = server_from_file(file_path)
    try:
        running_server = server_from_file(file_path, start_server=True)
    finally:
        running_server.stop()


@pytest.mark.unit
def test_yaml_client_loading():
    file_path = "yaml/konduit.yaml"
    client = client_from_file(file_path)


@pytest.mark.integration
def test_yaml_minimal_loading():
    file_path = "yaml/konduit_minimal.yaml"
    try:
        server = server_from_file(file_path, use_yaml=True, start_server=True)
        client = client_from_file(file_path, use_yaml=True)
    finally:
        server.stop()
    del server, client


@pytest.mark.integration
def test_json_minimal_loading():
    file_path = "yaml/konduit_minimal.json"
    try:
        server = server_from_file(file_path, use_yaml=False, start_server=True)
        client = client_from_file(file_path, use_yaml=False)
    finally:
        server.stop()
    del server, client


@pytest.mark.unit
def test_keras_serving():
    file_path = "yaml/konduit_keras.yaml"
    try:
        server = server_from_file(file_path=file_path)
    finally:
        server.stop()
    del server


@pytest.mark.unit
def test_tf_simple_serving():
    file_path = "yaml/konduit_tf_simple.yaml"
    try:
        server = server_from_file(file_path=file_path)
    finally:
        server.stop()
    del server


@pytest.mark.unit
def test_dl4j_mln_serving():
    file_path = "yaml/konduit_dl4j_mln.yaml"
    try:
        server = server_from_file(file_path=file_path)
    finally:
        server.stop()
    del server


@pytest.mark.unit
def test_dl4j_cg_serving():
    file_path = "yaml/konduit_dl4j_cg.yaml"
    try:
        server = server_from_file(file_path=file_path)
    finally:
        server.stop()
    del server


@pytest.mark.unit
def test_dl4j_samediff_serving():
    file_path = "yaml/konduit_samediff.yaml"
    try:
        server = server_from_file(file_path=file_path)
    finally:
        server.stop()
    del server


@pytest.mark.unit
def test_tensor_flow_serving():
    file_path = "yaml/konduit_tensorflow.yaml"
    try:
        server = server_from_file(file_path=file_path)
    finally:
        server.stop()
    del server


@pytest.mark.integration
def test_yaml_server_python_prediction():
    try:
        file_path = "yaml/konduit_python_code.yaml"
        server, client = konduit.load.from_file(file_path, start_server=True)
        client.predict(np.load("../data/input-0.npy"))
    finally:
        server.stop()


import pytest
from konduit.load import server_from_file
from konduit.load import store_pid, pop_pid
from konduit.server import stop_server_by_pid


@pytest.mark.unit
def test_pid_storage():
    file_path = "yaml/konduit.yaml"
    store_pid(file_path, 123)
    pid = pop_pid(file_path)
    assert pid == 123


@pytest.mark.integration
def test_pid_creation_removal():
    file_path = "yaml/konduit.yaml"
    running_server = server_from_file(file_path, start_server=True)

    # store the pid of this server and forget the Python object
    pid = running_server.process.pid
    store_pid(file_path=file_path, pid=running_server.process.pid)
    del running_server

    # retrieve the pid internally and kill the process
    recov_pid = pop_pid(file_path=file_path)
    assert pid == recov_pid
    stop_server_by_pid(recov_pid)



import os
import sys
import subprocess
from cli_init import KONDUIT_JAR_PATH


def cli():
    if not os.path.exists(KONDUIT_JAR_PATH):
        print("No konduit binaries found. See 'konduit-init --help' or just run 'konduit-init'"
              " to initialize a konduit jar binary.")
    else:
        arguments = ["java", "-jar", "-Dvertx.cli.usage.prefix=konduit", KONDUIT_JAR_PATH]
        arguments.extend(sys.argv[1:])

        subprocess.call(
            arguments,
            shell=sys.platform.startswith("win")
        )



def has_as_dict_attribute(python_object):
    """Checks if an object has a callable "as_dict" attribute

    :param python_object: any Python object
    """
    if not hasattr(python_object.__class__, "as_dict") and callable(
        getattr(python_object.__class__, "as_dict")
    ):
        raise AttributeError("Passed in Python object does not have an as_dict method.")


def _invoke_setter_on(input_object, method_name, value):
    """ If a given input object has the given setter method
    name, use this method to set "value" on the object.

    :param input_object: Python object
    :param method_name: name of the setter method
    :param value: value to set
    """
    remove = "_" + type(input_object).__name__ + "__"
    real_method_name = method_name.replace(remove, "")
    setter_method = "_set_" + real_method_name
    if hasattr(input_object, setter_method):
        method = getattr(input_object, setter_method)
        method(value)


def _ensure_serializable(input_config):
    """Modifies a given input configuration, if necessary,
    to make it serializable.
    """
    if not (
        hasattr(input_config, "__dict__")
        or type(input_config) is DictWrapper
        or type(input_config) is ListWrapper
    ):
        return

    # iterate over all config properties
    for setter_method, value in vars(input_config).items():
        if type(value) is dict:
            _invoke_setter_on(input_config, setter_method, DictWrapper(value))
        elif type(value) is list:
            for item in value:
                _ensure_serializable(item)
            _invoke_setter_on(input_config, setter_method, ListWrapper(value))

        elif hasattr(value, "as_dict"):
            _ensure_serializable(value)


def config_to_dict_with_type(inference_config):
    """
    Converts an inference configuration to a Python dictionary
    with 'type' key.

    :param inference_config: InferenceConfig object
    :return: Python dict
    """
    has_as_dict_attribute(inference_config)
    _ensure_serializable(inference_config)
    input_dict = inference_config.as_dict()
    input_dict["type"] = inference_config.__class__.__name__
    return input_dict


def empty_type_dict(input_object):
    """Generates an otherwise empty Python dict with the correct
    "type" key from Java.

    :param input_class: arbitrary instance of a Python class
    :return:
    """
    d = dict()
    d["type"] = input_object.__class__.__name__
    return d


class DictWrapper:
    """Wraps a Python dictionary. The `as_dict` method unrolls
    to a regular Python dict.
    """

    def __init__(self, input_dict):
        assert type(input_dict) is dict, "Input type must be a dictionary!"
        self.input_dict = input_dict

    def as_dict(self):
        ret = {}
        for key, value in self.input_dict.items():
            if hasattr(value, "as_dict"):
                ret[key] = value.as_dict()
            else:
                ret[key] = value

        return ret


class ListWrapper:
    """Wraps a Python list. The `as_dict` method should not
    be used. It's there for duck-type-style checks of Java
    classes.
    """

    def __init__(self, input_list):
        assert type(input_list) is list, "Input type must be a list!"
        self.input_list = input_list

    def as_dict(self):
        return self

    def as_list(self):
        return self.input_list

    def __iter__(self):
        return self.input_list.__iter__()



import io
import json
import logging
import numpy as np
import re
import requests
from konduit.utils import validate_server
from pyarrow.ipc import RecordBatchFileReader
from requests_toolbelt.multipart import decoder, encoder


class Client(object):
    def __init__(
        self,
        port,
        host="http://localhost",
        output_data_format=None,
        input_data_format=None,
        prediction_type=None,
        timeout=60,
        input_names=None,
        output_names=None,
    ):
        """Konduit Client

        This client is used to connect to a Konduit Server instance. You usually create a Client instance
        just by specifying host and port. If you want to convert the result of your endpoint to another data
        format on return, use `convert_to_format` to do so.

        :param port: The port on which the server is listening to. e.g. '1337' or 42, i.e. accepts int and str.
        :param host: The server host, defaults to 'http://localhost'.
        :param output_data_format: The output format returned by the `predict` method of this client. If not
               specified, we assume the output format is the same as the input format. This output format can
               be 'NUMPY', 'JSON', 'ARROW' and 'RAW'. 'IMAGE' will be implemented at a later stage.
        :param input_data_format: The format in which the input data is accepted by endpoints. Defaults to 'NUMPY',
               but can be 'JSON', 'ND4J', 'IMAGE' and 'ARROW' as well.
        :param prediction_type: The prediction type of the Konduit server. If not specified, this format
               will default to 'RAW'.
        :param timeout: Request time-out in seconds.
        :param input_names: The names of all inputs of the Konduit pipeline deployed for the Server corresponding to
               this client.
        :param output_names: The names of all inputs of the Konduit pipeline deployed for the Server corresponding to
               this client.
        """

        url = "{}:{}".format(host, port)

        server_running = validate_server(url)
        insufficient_data = output_data_format is None
        if server_running:
            try:
                response = requests.get("{}/config".format(url))
                config = response.json()
                logging.info("Retrieved config is".format(json.dumps(config)))
                steps = config["steps"]
                config = config["servingConfig"]
                input_names = []
                for step in steps:
                    input_names += step["inputNames"]
                if output_names is None:
                    output_names = []
                    for step in steps:
                        output_names += step["outputNames"]
                if output_data_format is None:
                    output_data_format = config["outputDataFormat"]
            except Exception as ex:
                logging.error(
                    "{}\nUnable to get configuration from the server. Please verify that the server is "
                    "running without any issues...".format(str(ex))
                )
                raise RuntimeError(ex)
        elif not server_running and insufficient_data:
            raise RuntimeError(
                "Unable to connect to the server at {}, not enough data provided to initialize "
                "the Client".format(url)
            )
        else:
            logging.info(
                "No server-side validation needed, enough information available to create Client."
            )

        if input_names is None:
            input_names = ["default"]
        assert isinstance(input_names, list), "Input names should be a list!"
        assert len(input_names) > 0, "Input names must not be empty!"

        if output_names is None:
            output_names = ["default"]
        assert isinstance(output_names, list), "Output names should be a list!"
        assert len(output_names) > 0, "Output names must not be empty!"

        if output_data_format:
            self.convert_to_format = output_data_format
        else:
            self.convert_to_format = input_data_format

        if input_data_format is None:
            input_data_format = "NUMPY"

        if prediction_type is None:
            prediction_type = "RAW"

        self.timeout = timeout
        self.input_format = input_data_format
        self.prediction_type = prediction_type
        self.input_names = input_names
        self.output_names = output_names
        self.url = url

    def predict(self, data_input=None):
        if isinstance(data_input, np.ndarray):
            data_input = {"default": data_input}
        if data_input is None:
            data_input = {}
        if self.input_format.upper() == "JSON":
            resp = requests.post(
                "{}/{}/{}".format(
                    self.url, self.prediction_type.lower(), self.input_format.lower()
                ),
                json=data_input,
                timeout=self.timeout,
            )

        else:
            self._validate_multi_part(data_input=data_input)
            converted_input = self._convert_multi_part_inputs(data_input=data_input)
            resp = requests.post(
                "{}/{}/{}".format(
                    self.url, self.prediction_type.lower(), self.input_format.lower()
                ),
                files=converted_input,
                timeout=self.timeout,
            )

        if resp.status_code != 200:
            raise Exception("Inference failed with error code: {}. "
                            "See the error log below for more info:\n---\n{}\n---\n"
                            .format(resp.status_code,
                                    str(resp.content, "utf-8")))
        else:
            if "content-type" not in resp.headers.keys():
                resp.headers["content-type"] = None
            return self._parse_response(resp.content, resp.headers["content-type"])

    def _parse_response(self, content, content_type):
        try:
            if self.output_names is None or len(self.output_names) < 2:
                bytes_content = io.BytesIO(content)
                bytes_content.seek(0)
                if self.convert_to_format.upper() == "NUMPY":
                    return np.load(bytes_content)
                elif self.convert_to_format.upper() == "ARROW":
                    reader = RecordBatchFileReader(bytes_content)
                    return reader.read_pandas()
                elif self.convert_to_format.upper() == "JSON":
                    return json.load(bytes_content)
                elif self.convert_to_format.upper() == "IMAGE":
                    raise NotImplementedError("Image not implemented yet.")
                elif self.convert_to_format.upper() == "RAW":
                    return content
            else:
                return self._convert_multi_part_output(content, content_type)
        except Exception as e:
            logging.info(e)
            return None

    @staticmethod
    def _convert_numpy_to_binary(input_arr):
        bytes_io = io.BytesIO()
        np.save(bytes_io, input_arr)
        bytes_io.seek(0)
        return bytes_io

    @staticmethod
    def _convert_binary_to_numpy(input_arr_bytes):
        bytes_io = io.BytesIO(input_arr_bytes)
        ret = np.load(bytes_io)
        return ret

    @staticmethod
    def _encode_multi_part_input(parts=None):
        if parts is None:
            parts = {}
        encoded_parts = [(k, v) for k, v in parts.items()]
        ret = encoder.MultipartEncoder(encoded_parts)
        return ret.to_string(), ret.content_type

    @staticmethod
    def _convert_multi_part_inputs(data_input=None):
        if data_input is None:
            data_input = {}
        ret = {}
        for key, value in data_input.items():
            if isinstance(value, np.ndarray):
                value = Client._convert_numpy_to_binary(value)
            ret[key] = value
        return ret

    def _convert_multi_part_output(self, content, content_type):
        multipart_data = decoder.MultipartDecoder(
            content=content, content_type=content_type
        )
        ret = {}
        for part in multipart_data.parts:
            # typically something like: b'form-data; name="input1"'
            name_with_form_data = str(part.headers[b"Content-Disposition"])
            name_str = re.sub(
                r'([";\\\']|name=|form-data|b\\)', "", name_with_form_data
            ).replace("b ", "")
            if self.prediction_type.upper() == "NUMPY":
                ret[name_str] = Client._convert_binary_to_numpy(part.content)
            elif self.prediction_type.upper() == "ARROW":
                reader = RecordBatchFileReader(part.content)
                ret[name_str] = reader.read_pandas()
            elif self.prediction_type.upper() == "JSON":
                return json.load(part.content)
            elif self.prediction_type.upper() == "ND4J":
                raise NotImplementedError("Nd4j not implemented yet.")
            else:
                ret[name_str] = part.content
        return ret

    def _validate_multi_part(self, data_input):
        if self.input_format.capitalize() == "JSON":
            raise ValueError(
                "Attempting to execute multi part request with input type specified as JSON."
            )

        for key, value in data_input.items():
            root_name = re.sub("\\[[0-9]+\\]", "", key)
            if root_name not in self.input_names:
                raise ValueError(
                    "Specified root name " + root_name + " not found in input names."
                )



import pytest
from konduit import *


@pytest.mark.integration
def test_client_from_server():
    server_id = "python_server"

    python_config = PythonConfig(
        python_code="first += 2",
        python_inputs={"first": "NDARRAY"},
        python_outputs={"first": "NDARRAY"},
    )

    step = PythonStep().step(python_config)
    server = Server(steps=step, serving_config=ServingConfig())
    server.start(server_id)

    try:
        server.get_client()
    finally:
        server.stop(server_id)


@pytest.mark.unit
def test_multipart_regex():
    client = Client(
        port=1337,
        output_data_format="NUMPY",
        input_names=["partname"],
        output_names=["nobody_cares"],
    )

    test_data = {"partname[0]": "foo", "partname[1]": "bar"}

    client._validate_multi_part(test_data)

    test_data["foo"] = "baz"

    with pytest.raises(Exception):
        client._validate_multi_part(test_data)



import pytest
from konduit import *
from konduit.load import server_from_file, client_from_file
import numpy as np

@pytest.mark.integration
def test_wordpiece_tokenizer_serving_minimal():

    file_path = "yaml/konduit_wordpiece_tokenizer_minimal.yaml"
    server = server_from_file(file_path)
    try:
        running_server = server_from_file(file_path, start_server=True)
    finally:
        running_server.stop()


@pytest.mark.integration
def test_wordpiece_tokenizer_serving_two_steps():

    file_path = "yaml/konduit_wordpiece_tokenizer_two_steps.yaml"
    server = server_from_file(file_path)
    try:
        running_server = server_from_file(file_path, start_server=True)
    finally:
        running_server.stop()



import json
import numpy as np
import pytest
import random
from konduit import *
from konduit.client import Client
from konduit.json_utils import config_to_dict_with_type


@pytest.mark.unit
def test_multipart_encode():
    input_names = ["IteratorGetNext:0", "IteratorGetNext:1", "IteratorGetNext:4"]
    output_names = ["loss/Softmax"]
    port = random.randint(1000, 65535)
    client = Client(
        input_names=input_names,
        output_names=output_names,
        input_data_format="NUMPY",
        output_data_format="NUMPY",
        port=port,
    )

    input_data = {
        "input1": Client._convert_numpy_to_binary(np.ones(1)),
        "input2": Client._convert_numpy_to_binary(np.ones(2)),
    }

    converted = Client._convert_multi_part_inputs(input_data)
    body, content_type = Client._encode_multi_part_input(converted)

    client._convert_multi_part_output(content=body, content_type=content_type)


@pytest.mark.unit
def test_python_serde():
    python_configuration = PythonConfig(
        python_code="first += 2", python_inputs=["first"], python_outputs=["first"]
    )

    port = random.randint(1000, 65535)
    serving_config = ServingConfig(
        http_port=port,
        output_data_format="NUMPY",
        log_timings=True,
    )

    python_pipeline_step = PythonStep().step(python_configuration)
    inference_config = InferenceConfiguration(
        serving_config=serving_config, steps=[python_pipeline_step]
    )

    json.dumps(config_to_dict_with_type(inference_config))



import pytest
from konduit import *
from konduit.json_utils import *


@pytest.mark.unit
def test_empty_dict_type():
    d1 = empty_type_dict(InferenceConfiguration())
    d2 = {"type": "InferenceConfiguration"}
    assert d1 == d2


@pytest.mark.unit
def test_dict_wrapper():
    d = {"foo": {"bar": "baz"}}
    dw = DictWrapper(d)

    assert d == dw.as_dict()


@pytest.mark.unit
def test_list_wrapper():
    lst = ["a", "b", "c"]
    lw = ListWrapper(lst)

    # This is a national tragedy, but OK
    assert lw == lw.as_dict()

    assert lst == lw.as_list()

    x = ""
    for i in lw:
        x = i
    assert x == "c"


@pytest.mark.unit
def test_as_dict_checker():
    class FooCallable(object):
        def __init__(self):
            pass

        def as_dict(self):
            pass

    class FooNotCallable(object):
        def __init__(self):
            self.as_dict = {}

    has_as_dict_attribute(FooCallable())

    with pytest.raises(Exception):
        has_as_dict_attribute(FooNotCallable())



import json
import pytest
from jnius import autoclass
from konduit import *
from konduit.json_utils import config_to_dict_with_type

StringJava = autoclass("java.lang.String")
InferenceConfigurationJava = autoclass("ai.konduit.serving.InferenceConfiguration")


@pytest.mark.unit
@pytest.mark.skip("Can't seem to find the toJson method inside the inference configuration. This might be due to the "
                  "TextConfig interface addition")
# TODO: fix this test
def test_json_compare():
    parallel_inference_config = ParallelInferenceConfig(workers=1)
    serving_config = ServingConfig(
        http_port=1300, output_data_format="NUMPY"
    )

    model_pipeline_step = TensorFlowStep(
        path="model.pb",
        parallel_inference_config=parallel_inference_config,
        input_names=["IteratorGetNext:0", "IteratorGetNext:1", "IteratorGetNext:4"],
        output_names=["loss/Softmax"],
    )

    inference = InferenceConfiguration(
        serving_config=serving_config, steps=[model_pipeline_step]
    )

    assert_config_works(inference)


@pytest.mark.unit
@pytest.mark.skip("Can't seem to find the toJson method inside the inference configuration. This might be due to the "
                  "TextConfig interface addition")
# TODO: fix this test
def test_python_serde():
    input_names = ["default"]
    output_names = ["default"]

    python_config = PythonConfig(
        python_code="first += 2",
        python_inputs={"first": "NDARRAY"},
        python_outputs={"first": "NDARRAY"},
    )

    port = 1300
    serving_config = ServingConfig(
        http_port=port,
        output_data_format="NUMPY",
        log_timings=True,
    )

    python_pipeline_step = PythonStep(
        input_names=input_names,
        output_names=output_names,
        python_configs=DictWrapper({"default": python_config}),
    )

    inference = InferenceConfiguration(
        serving_config=serving_config, steps=[python_pipeline_step]
    )

    assert_config_works(inference)


def assert_config_works(config):
    config_json = config_to_dict_with_type(config)
    json_str = json.dumps(config_json)
    config = InferenceConfigurationJava.fromJson(StringJava(json_str))
    test_json = config.toJson()
    test_json_dict = json.loads(test_json)
    config_json_dict = config_json
    test_pipeline_steps = test_json_dict["steps"]
    config_pipeline_steps = config_json_dict["steps"]
    assert len(test_pipeline_steps) == len(config_pipeline_steps)


from setuptools import find_packages
from setuptools import setup

setup(
    name="konduit",
    version="0.1.10",
    packages=find_packages(),
    install_requires=[
        "requests>=2.22.0",
        "numpy<=1.16.4",  # For compatibility with python 2
        "pyarrow==0.13.0",
        "requests-toolbelt>=0.9.1",
        "pandas<=0.24.2",  # For compatibility with python 2
        "Cython",
        "pydl4j",
        "pydatavec",
        "pyyaml",
        "click",
        "packaging",
        "hurry.filesize"
    ],
    py_modules=["konduit", "cli", "cli_init"],
    extras_require={
        "tests": ["pytest", "pytest-pep8", "pytest-cov", "mock"],
        "codegen": ["jsonschema2popo"],
        "dev": ["black", "pre-commit"],
    },
    entry_points={"console_scripts": ["konduit=cli:cli", "konduit-init=cli_init:init"]},
    include_package_data=True,
    license="Apache",
    description="konduit: Enterprise Runtime for Machine Learning Models",
    long_description="konduit: Enterprise Runtime for Machine Learning Models",
    author="Max Pumperla, Shams Ul Azeem",
    author_email="max@konduit.ai, shams@konduit.ai",
    url="https://github.com/KonduitAI/konduit-serving",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)



import glob
import pytest
import re
from konduit import *


def markdown_blocks(file_path, language="python"):
    """Get language-specific markdown code blocks.
    :param file_path: path to markdown file
    :param language: 'python', 'java' (or literally any valid code fence language)
    :return: list of code as string
    """
    code_blocks = []
    code_regex = r"^```.*"
    code_block_open_re = r"^```(`*)({0})$".format(language)

    with open(file_path, "r") as f:
        block = []
        python = True
        in_code_block = False

        for line in f.readlines():
            code_block_delimiter = re.match(code_regex, line)

            if in_code_block:
                if code_block_delimiter:
                    if python:
                        code_blocks.append("".join(block))
                    block = []
                    python = True
                    in_code_block = False
                else:
                    block.append(line)
            elif code_block_delimiter:
                in_code_block = True
                if not re.match(code_block_open_re, line):
                    python = False
    return code_blocks


def is_markdown(f):
    """Does this file have a markdown extension?"""
    markdown_extensions = [".markdown", ".mdown", ".mkdn", ".mkd", ".md"]
    return os.path.splitext(f)[1] in markdown_extensions


def get_files(input_dir):
    """Get all markdown files in a directory recursively."""
    return [
        os.path.join(dp, f)
        for dp, dn, file_names in os.walk(input_dir)
        for f in file_names
        if is_markdown(f)
    ]


def exec_python_code(code_blocks):
    """Execute the Python code contained in blocks.

    :param code_blocks: List of strings containing executable Python code
    """
    for block in code_blocks:
        try:
            exec(block)
        except Exception as ex:
            print("Test execution failed in the following block:\n")
            print(block)
            print("failed with error message:\n")
            raise ex


def make_folders():
    """Make base folders for Java application logic."""
    if not os.path.isdir("./ai/konduit/serving"):
        os.makedirs("./ai/konduit/serving")


def is_self_contained_example(block):
    """Basic sanity check if markdown block contains a class and a main method."""
    return "public static void main" in block and "public class" in block


def write_self_contained_example(block):
    """Write the full code block to a file if we're dealing with a self-contained
    example.

    :param block: code block as string
    """
    class_regex = r"public\s+class\s+(\w+)"
    class_name = re.search(class_regex, block)
    if class_name:
        class_name = class_name.group(1)
        with open("ai/konduit/serving/" + class_name + ".java", "w") as f:
            f.write(block)
    else:
        raise Exception("Could not determine proper class name")


def write_example_from_snippet(block, markdown_root, i):
    """Write imports first and supply main code snippets to a method in a constructed Java class
    :param block: code block
    :param markdown_root: name of the file this snippet comes from (without extension)
    :param i: this block is the i-th example  in the current markdown file
    """
    lines = block.split("\n")
    import_lines = [l for l in lines if l.startswith("import ")]
    code_lines = [l for l in lines if not l.startswith("import ")]
    class_name = "BasicsTest" + markdown_root + str(i)

    code = (
        "package ai.konduit.serving;\n\n"
        + "\n".join(import_lines)
        + "\n\npublic class "
        + class_name
        + " {\n"
        + "\tpublic "
        + class_name
        + " () {}\n"
        + "\tpublic void main() {\n\t\t"
        + "\n\t\t".join(code_lines)
        + "\n\t}\n}"
    )

    with open("ai/konduit/serving/" + class_name + ".java", "w") as f:
        f.write(code)


def get_markdown_root(file_path):
    full_path = os.path.splitext(file_path)[0]
    return full_path.split(os.sep)[-1]


def write_java_files(code_blocks, file_path):
    """Write code contained in blocks into proper Java source files.
    We cover two basic scenarios:
        1. the code is a self-contained Java app with main method
        2. we just have a small snippet of imports and statements (that run when correctly
           (transformed into a class)

    :param code_blocks: code blocks as list of string
    :param file_path: path to file containing the code
    """
    markdown_root = get_markdown_root(file_path)
    make_folders()
    for i, block in enumerate(code_blocks):
        if is_self_contained_example(block):
            write_self_contained_example(block)
        else:
            write_example_from_snippet(block, markdown_root, i)


def exec_java_code(code_blocks, file_path):
    """Execute the java classes previously put into the konduit.jar by the
    `prepare_doc_tests.sh` script.

    :param code_blocks: code blocks as list of string
    :param file_path: path to file containing the code
    :return:
    """
    markdown_root = get_markdown_root(file_path)

    for i, block in enumerate(code_blocks):
        if is_self_contained_example(block):
            class_regex = r"public\s+class\s+(\w+)"
            class_name = re.search(class_regex, block).group(1)
        else:
            class_name = "BasicsTest{}{}".format(markdown_root, str(i))
        basic_test_class = autoclass("ai.konduit.serving." + class_name)
        basic_test = basic_test_class()
        if is_self_contained_example(block):
            basic_test.main([])
        else:
            basic_test.main()


def clean_java_files(ext=""):
    """Clean Java source files and classes"""
    for file_name in glob.glob("./ai/konduit/serving/*" + ext):
        try:
            os.remove(file_name)
        except FileNotFoundError:
            continue


@pytest.mark.unit
def test_docs():
    """This is the main unit test for testing documentation code snippets
    contained in markdown files."""
    files = get_files("../docs")
    for file_path in files:
        python_blocks = markdown_blocks(file_path, "python./")
        exec_python_code(python_blocks)

        java_blocks = markdown_blocks(file_path, "java")
        exec_java_code(java_blocks, file_path)
        clean_java_files()


def _prepare_docs_jar():
    """This helper will be called from the `prepare_doc_tests.sh` script."""
    files = get_files("../docs")
    for file_path in files:
        java_blocks = markdown_blocks(file_path, "java")
        write_java_files(java_blocks, file_path)


if __name__ == "__main__":
    print("prepare java files")
    _prepare_docs_jar()



import logging
import os
import requests
import socket
import sys
from contextlib import closing


def is_port_in_use(port, base_url="localhost"):
    """is this port already in use for a Konduit server?"""
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        return s.connect_ex((base_url, port)) == 0


def default_python_path(work_dir="."):
    """set and return the Python path by appending this work_dir to it."""
    work_dir = os.path.abspath(work_dir)
    python_path = os.pathsep.join(sys.path)
    python_path += os.pathsep + work_dir
    return python_path


def validate_server(url):
    """
    Validates if a konduit.Server is running under the specified url. Returns True if
    the server is running, False otherwise.

    :param url: host and port of the server as str
    :return: boolean
    """
    try:
        r = requests.get("{}/healthcheck".format(url))
        if r.status_code != 204:
            logging.error(
                "The server health checks failed. Please verify that the server is running without any "
                "issues..."
            )
        else:
            return True
    except Exception as ex:
        logging.error(
            "{}\nUnable to connect to the server or the server health checks have failed. Please "
            "verify that the server is running without any issues...".format(str(ex))
        )
        return False


def to_unix_path(file_path):
    """Utility function to turn a Windows-style path into Unix-style. Windows
    can handle the latter, so we use this representation internally. Python can
    handle this in any case. We also replace Windows' ";" for path separation by
    os.pathsep to make sure this runs on every system.

    :param file_path: path to your file
    :return: Unix-style version of your file
    """
    return file_path.replace("\\", "/").replace(";", os.pathsep)


def update_dict_with_unix_paths(config):
    """Replace all str values in Python dictionary that correspond to paths
    by their Unix-style equivalent.

    :param config: any Python dictionary
    :return: updated dict
    """
    for k, v in config.items():
        if "_path" in k and isinstance(v, str):
            config[k] = to_unix_path(v)
    return config



import numpy as np
import pytest
from konduit import (
    ParallelInferenceConfig,
    ServingConfig,
    TensorFlowStep,
    InferenceConfiguration
)
from konduit.client import Client
from konduit.server import Server
from konduit.utils import is_port_in_use


@pytest.mark.integration
def test_server_start():
    server_id = "tensorflow_server"
    input_names = ["IteratorGetNext:0", "IteratorGetNext:1", "IteratorGetNext:4"]
    output_names = ["loss/Softmax"]
    parallel_inference_config = ParallelInferenceConfig(workers=1)
    serving_config = ServingConfig(
        output_data_format="NUMPY",
        log_timings=True,
    )

    model_pipeline_step = TensorFlowStep(
        path="bert_mrpc_frozen.pb",
        input_data_types={
            input_names[0]: "INT32",
            input_names[1]: "INT32",
            input_names[2]: "INT32",
        },
        parallel_inference_config=parallel_inference_config,
        input_names=input_names,
        output_names=output_names,
    )

    inference = InferenceConfiguration(
        serving_config=serving_config, steps=[model_pipeline_step]
    )

    server = Server(
        inference_config=inference, extra_start_args="-Xmx8g"
    )
    _, port, started = server.start(server_id)

    data_input = {
        input_names[0]: np.load("../data/input-0.npy"),
        input_names[1]: np.load("../data/input-1.npy"),
        input_names[2]: np.load("../data/input-4.npy"),
    }

    assert started  # will be true if the server was started
    assert is_port_in_use(port)

    client = Client(input_data_format="NUMPY", prediction_type="RAW", port=port)

    try:
        predicted = client.predict(data_input)
        print(predicted)
    except Exception as e:
        print(e)
    finally:
        server.stop(server_id)
